@@ -25,9 +25,9 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <libgnomevfs/gnome-vfs.h>
 
 #include "menu-util.h"
-#include "menu-monitor.h"
 #include "canonicalize.h"
 
 typedef struct CachedDir CachedDir;
@@ -58,13 +58,11 @@ struct CachedDir
   GSList *entries;
   GSList *subdirs;
 
-  MenuMonitor *dir_monitor;
-  GSList      *monitors;
+  GnomeVFSMonitorHandle *monitor;
+  GSList                *monitors;
 
   guint have_read_entries : 1;
-  guint deleted : 1;
-
-  guint references : 28;
+  guint no_monitor_support : 1;
 };
 
 struct CachedDirMonitor
@@ -74,17 +72,10 @@ struct CachedDirMonitor
   gpointer                   user_data;
 };
 
-static void     cached_dir_add_reference          (CachedDir *dir);
-static void     cached_dir_remove_reference       (CachedDir *dir);
 static void     cached_dir_free                   (CachedDir  *dir);
 static gboolean cached_dir_load_entries_recursive (CachedDir  *dir,
                                                    const char *dirname);
 
-static void handle_cached_dir_changed (MenuMonitor      *monitor,
-				       MenuMonitorEvent  event,
-				       const char       *path,
-				       CachedDir        *dir);
-
 /*
  * Entry directory cache
  */
@@ -104,32 +95,38 @@ cached_dir_new (const char *name)
 }
 
 static void
-cached_dir_free (CachedDir *dir)
+cached_dir_clear_entries (CachedDir *dir)
 {
-  if (dir->dir_monitor)
-    {
-      menu_monitor_remove_notify (dir->dir_monitor,
-				  (MenuMonitorNotifyFunc) handle_cached_dir_changed,
-				  dir);
-      menu_monitor_unref (dir->dir_monitor);
-      dir->dir_monitor = NULL;
-    }
-
-  g_slist_foreach (dir->monitors, (GFunc) g_free, NULL);
-  g_slist_free (dir->monitors);
-  dir->monitors = NULL;
-
   g_slist_foreach (dir->entries,
                    (GFunc) desktop_entry_unref,
                    NULL);
   g_slist_free (dir->entries);
   dir->entries = NULL;
+}
 
+static void
+cached_dir_clear_subdirs (CachedDir *dir)
+{
   g_slist_foreach (dir->subdirs,
                    (GFunc) cached_dir_free,
                    NULL);
   g_slist_free (dir->subdirs);
   dir->subdirs = NULL;
+}
+
+static void
+cached_dir_free (CachedDir *dir)
+{
+  cached_dir_clear_entries (dir);
+  cached_dir_clear_subdirs (dir);
+
+  if (dir->monitor)
+    gnome_vfs_monitor_cancel (dir->monitor);
+  dir->monitor = NULL;
+
+  g_slist_foreach (dir->monitors, (GFunc) g_free, NULL);
+  g_slist_free (dir->monitors);
+  dir->monitors = NULL;
 
   g_free (dir->name);
   g_free (dir);
@@ -186,6 +183,9 @@ cached_dir_find_relative_path (CachedDir  *dir,
   i = 0;
   while (split[i] != NULL)
     {
+      if (*(split[i]) == '\0')
+        continue;
+
       if (split[i + 1] != NULL)
         {
           if ((dir = find_subdir (dir, split[i])) == NULL)
@@ -205,6 +205,74 @@ cached_dir_find_relative_path (CachedDir  *dir,
   return retval;
 }
 
+static DesktopEntry *
+cached_dir_find_file_id (CachedDir  *dir,
+                         const char *file_id,
+                         gboolean    legacy)
+{
+  DesktopEntry *retval = NULL;
+
+  if (!legacy)
+    {
+      char *p;
+      char *freeme;
+
+      p = freeme = g_strdup (file_id);
+      while (p != NULL)
+        {
+          char *q;
+
+          if ((retval = find_entry (dir, p)))
+            break;
+
+          q = strchr (p, '-');
+          if (q)
+            {
+              CachedDir *subdir;
+
+              *(q++) = '\0';
+
+              if ((subdir = find_subdir (dir, p)))
+                {
+                  retval = cached_dir_find_file_id (subdir, q, legacy);
+                  if (retval)
+                    break;
+                }
+            }
+
+          p = q;
+        }
+
+      g_free (freeme);
+    }
+  else /* legacy */
+    {
+      retval = find_entry (dir, file_id);
+
+      /* Ignore entries with categories in legacy trees
+       */
+      if (retval && desktop_entry_has_categories (retval))
+        retval = NULL;
+
+      if (!retval)
+        {
+          GSList *tmp;
+
+          tmp = dir->subdirs;
+          while (tmp != NULL)
+            {
+              retval = cached_dir_find_file_id (tmp->data, file_id, legacy);
+              if (retval)
+                break;
+
+              tmp = tmp->next;
+            }
+        }
+    }
+
+  return retval;
+}
+
 static CachedDir *
 cached_dir_lookup (const char *canonical)
 {
@@ -216,7 +284,7 @@ cached_dir_lookup (const char *canonical)
     dir_cache = cached_dir_new ("/");
   dir = dir_cache;
 
-  g_assert (canonical != NULL && canonical[0] == G_DIR_SEPARATOR);
+  g_assert (canonical != NULL && canonical[0] == '/');
 
   menu_verbose ("Looking up cached dir \"%s\"\n", canonical);
 
@@ -227,6 +295,9 @@ cached_dir_lookup (const char *canonical)
     {
       CachedDir *subdir;
 
+      if (*(split[i]) == '\0')
+        continue;
+
       if ((subdir = find_subdir (dir, split[i])) == NULL)
         {
           subdir = cached_dir_new (split[i]);
@@ -276,6 +347,7 @@ cached_dir_update_entry (CachedDir  *dir,
         {
           if (!desktop_entry_reload (tmp->data))
 	    {
+	      desktop_entry_unref (tmp->data);
 	      dir->entries = g_slist_delete_link (dir->entries, tmp);
 	    }
 
@@ -317,14 +389,6 @@ cached_dir_add_subdir (CachedDir  *dir,
 {
   CachedDir *subdir;
 
-  subdir = find_subdir (dir, basename);
-
-  if (subdir != NULL)
-    {
-      subdir->deleted = FALSE;
-      return TRUE;
-    }
-
   subdir = cached_dir_new (basename);
 
   if (!cached_dir_load_entries_recursive (subdir, path))
@@ -345,144 +409,67 @@ static gboolean
 cached_dir_remove_subdir (CachedDir  *dir,
                           const char *basename)
 {
-  CachedDir *subdir;
+  GSList *tmp;
 
-  subdir = find_subdir (dir, basename);
-
-  if (subdir != NULL)
+  tmp = dir->subdirs;
+  while (tmp != NULL)
     {
-      subdir->deleted = TRUE;
+      CachedDir *subdir = tmp->data;
 
-      if (subdir->references == 0)
+      if (strcmp (subdir->name, basename) == 0)
         {
-          cached_dir_free (subdir);
-          dir->subdirs = g_slist_remove (dir->subdirs, subdir);
+	  cached_dir_free (tmp->data);
+          dir->subdirs = g_slist_delete_link (dir->subdirs, tmp);
+          return TRUE;
         }
 
-      return TRUE;
-    }
-
-  return FALSE;
-}
-
-static guint   monitors_idle_handler = 0;
-static GSList *pending_monitors_dirs = NULL;
-
-static void
-cached_dir_invoke_monitors (CachedDir *dir)
-{
-  GSList *tmp;
-
-  tmp = dir->monitors;
-  while (tmp != NULL)
-    {
-      CachedDirMonitor *monitor = tmp->data;
-      GSList           *next    = tmp->next;
-
-      monitor->callback (monitor->ed, monitor->user_data);
-
-      tmp = next;
-    }
-
-  /* we explicitly don't invoke monitors of the parent since an
-   * event has been queued for it too */
-}
-
-static gboolean
-emit_monitors_in_idle (void)
-{
-  GSList *monitors_to_emit;
-  GSList *tmp;
-
-  monitors_to_emit = pending_monitors_dirs;
-
-  pending_monitors_dirs = NULL;
-  monitors_idle_handler = 0;
-
-  tmp = monitors_to_emit;
-  while (tmp != NULL)
-    {
-      CachedDir *dir = tmp->data;
-
-      cached_dir_invoke_monitors (dir);
-      cached_dir_remove_reference (dir);
-
       tmp = tmp->next;
     }
 
-  g_slist_free (monitors_to_emit);
-
   return FALSE;
 }
 
 static void
-cached_dir_queue_monitor_event (CachedDir *dir)
-{
-  GSList *tmp;
-
-  tmp = pending_monitors_dirs;
-  while (tmp != NULL)
-    {
-      CachedDir *d    = tmp->data;
-      GSList    *next = tmp->next;
-
-      if (dir->parent == d->parent &&
-          g_strcmp0 (dir->name, d->name) == 0)
-        break;
-
-      tmp = next;
-    }
-
-  /* not found, so let's queue it */
-  if (tmp == NULL)
-    {
-      cached_dir_add_reference (dir);
-      pending_monitors_dirs = g_slist_append (pending_monitors_dirs, dir);
-    }
-
-  if (dir->parent)
-    {
-      cached_dir_queue_monitor_event (dir->parent);
-    }
-
-  if (monitors_idle_handler == 0)
-    {
-      monitors_idle_handler = g_idle_add ((GSourceFunc) emit_monitors_in_idle, NULL);
-    }
-}
-
-static void
-handle_cached_dir_changed (MenuMonitor      *monitor,
-			   MenuMonitorEvent  event,
-			   const char       *path,
-                           CachedDir        *dir)
+handle_cached_dir_changed (GnomeVFSMonitorHandle    *handle,
+                           const char               *monitor_uri,
+                           const char               *info_uri,
+                           GnomeVFSMonitorEventType  event,
+                           CachedDir                *dir)
 {
   gboolean  handled = FALSE;
+  char     *path;
   char     *basename;
-  char     *dirname;
 
-  menu_verbose ("'%s' notified of '%s' %s - invalidating cache\n",
-		dir->name,
+  if (event != GNOME_VFS_MONITOR_EVENT_CREATED &&
+      event != GNOME_VFS_MONITOR_EVENT_DELETED &&
+      event != GNOME_VFS_MONITOR_EVENT_CHANGED)
+    return;
+
+  if (!(path = gnome_vfs_get_local_path_from_uri (info_uri)))
+    return;
+
+  menu_verbose ("Notified of '%s' %s - invalidating cache\n",
                 path,
-                event == MENU_MONITOR_EVENT_CREATED ? ("created") :
-                event == MENU_MONITOR_EVENT_DELETED ? ("deleted") : ("changed"));
+                event == GNOME_VFS_MONITOR_EVENT_CREATED ? ("created") :
+                event == GNOME_VFS_MONITOR_EVENT_DELETED ? ("deleted") :
+                event == GNOME_VFS_MONITOR_EVENT_CHANGED ? ("changed") : ("unknown-event"));
 
-  dirname  = g_path_get_dirname  (path);
   basename = g_path_get_basename (path);
 
-  dir = cached_dir_lookup (dirname);
-
   if (g_str_has_suffix (basename, ".desktop") ||
       g_str_has_suffix (basename, ".directory"))
     {
       switch (event)
         {
-        case MENU_MONITOR_EVENT_CREATED:
-        case MENU_MONITOR_EVENT_CHANGED:
+        case GNOME_VFS_MONITOR_EVENT_CREATED:
+          handled = cached_dir_add_entry (dir, basename, path);
+          break;
+
+        case GNOME_VFS_MONITOR_EVENT_CHANGED:
           handled = cached_dir_update_entry (dir, basename, path);
           break;
 
-        case MENU_MONITOR_EVENT_DELETED:
+        case GNOME_VFS_MONITOR_EVENT_DELETED:
           handled = cached_dir_remove_entry (dir, basename);
           break;
 
@@ -495,14 +482,14 @@ handle_cached_dir_changed (MenuMonitor      *monitor,
     {
       switch (event)
         {
-        case MENU_MONITOR_EVENT_CREATED:
+        case GNOME_VFS_MONITOR_EVENT_CREATED:
           handled = cached_dir_add_subdir (dir, basename, path);
           break;
 
-        case MENU_MONITOR_EVENT_CHANGED:
+        case GNOME_VFS_MONITOR_EVENT_CHANGED:
           break;
 
-        case MENU_MONITOR_EVENT_DELETED:
+        case GNOME_VFS_MONITOR_EVENT_DELETED:
           handled = cached_dir_remove_subdir (dir, basename);
           break;
 
@@ -513,17 +500,22 @@ handle_cached_dir_changed (MenuMonitor      *monitor,
     }
 
   g_free (basename);
-  g_free (dirname);
+  g_free (path);
 
   if (handled)
     {
-      /* CHANGED events don't change the set of desktop entries */
-      if (event == MENU_MONITOR_EVENT_CREATED || event == MENU_MONITOR_EVENT_DELETED)
-        {
-          _entry_directory_list_empty_desktop_cache ();
-        }
+      GSList *tmp;
 
-      cached_dir_queue_monitor_event (dir);
+      tmp = dir->monitors;
+      while (tmp != NULL)
+        {
+          CachedDirMonitor *monitor = tmp->data;
+	  GSList           *next    = tmp->next;
+
+          monitor->callback (monitor->ed, monitor->user_data);
+
+          tmp = next;
+        }
     }
 }
 
@@ -531,13 +523,23 @@ static void
 cached_dir_ensure_monitor (CachedDir  *dir,
                            const char *dirname)
 {
-  if (dir->dir_monitor == NULL)
+  char *uri;
+
+  if (dir->monitor != NULL || dir->no_monitor_support)
+    return;
+
+  uri = gnome_vfs_get_uri_from_local_path (dirname);
+
+  if (gnome_vfs_monitor_add (&dir->monitor,
+                             uri,
+                             GNOME_VFS_MONITOR_DIRECTORY,
+                             (GnomeVFSMonitorCallback) handle_cached_dir_changed,
+                             dir) != GNOME_VFS_OK)
     {
-      dir->dir_monitor = menu_get_directory_monitor (dirname);
-      menu_monitor_add_notify (dir->dir_monitor,
-			       (MenuMonitorNotifyFunc) handle_cached_dir_changed,
-			       dir);
+      dir->no_monitor_support = TRUE;
     }
+
+  g_free (uri);
 }
 
 static gboolean
@@ -565,11 +567,12 @@ cached_dir_load_entries_recursive (CachedDir  *dir,
       return FALSE;
     }
 
+  cached_dir_clear_entries (dir);
   cached_dir_ensure_monitor (dir, dirname);
 
   fullpath = g_string_new (dirname);
-  if (fullpath->str[fullpath->len - 1] != G_DIR_SEPARATOR)
-    g_string_append_c (fullpath, G_DIR_SEPARATOR);
+  if (fullpath->str[fullpath->len - 1] != '/')
+    g_string_append_c (fullpath, '/');
 
   fullpath_len = fullpath->len;
 
@@ -606,6 +609,52 @@ cached_dir_load_entries_recursive (CachedDir  *dir,
   return TRUE;
 }
 
+static char *
+cached_dir_get_full_path (CachedDir *dir)
+{
+  GString *str;
+  GSList  *parents = NULL;
+  GSList  *tmp;
+
+  if (dir == NULL || dir->parent == NULL)
+    return g_strdup ("/");
+
+  while (dir->parent != NULL)
+    {
+      parents = g_slist_prepend (parents, dir->name);
+      dir = dir->parent;
+    }
+
+  str = g_string_new ("/");
+
+  tmp = parents;
+  while (tmp != NULL)
+    {
+      g_string_append (str, tmp->data);
+      g_string_append_c (str, '/');
+
+      tmp = tmp->next;
+    }
+
+  g_slist_free (parents);
+
+  return g_string_free (str, FALSE);
+}
+
+static CachedDir *
+cached_dir_load (const char *canonical_path)
+{
+  CachedDir *retval;
+
+  menu_verbose ("Loading cached dir \"%s\"\n", canonical_path);
+
+  retval = cached_dir_lookup (canonical_path);
+
+  cached_dir_load_entries_recursive (retval, canonical_path);
+
+  return retval;
+}
+
 static void
 cached_dir_add_monitor (CachedDir                 *dir,
                         EntryDirectory            *ed,
@@ -666,51 +715,32 @@ cached_dir_remove_monitor (CachedDir                 *dir,
 }
 
 static void
-cached_dir_add_reference (CachedDir *dir)
+cached_dir_ensure_loaded (CachedDir *dir)
 {
-  dir->references++;
+  char *path;
 
-  if (dir->parent != NULL)
-    {
-      cached_dir_add_reference (dir->parent);
-    }
+  if (dir->have_read_entries)
+    return;
+
+  path = cached_dir_get_full_path (dir);
+  cached_dir_load_entries_recursive (dir, path);
+  g_free (path);
 }
 
-static void
-cached_dir_remove_reference (CachedDir *dir)
+static GSList*
+cached_dir_get_subdirs (CachedDir *dir)
 {
-  CachedDir *parent;
+  cached_dir_ensure_loaded (dir);
 
-  parent = dir->parent;
+  return dir->subdirs;
+}
 
-  if (--dir->references == 0 && dir->deleted)
-    {
-      if (dir->parent != NULL)
-	{
-	  GSList *tmp;
+static GSList*
+cached_dir_get_entries (CachedDir   *dir)
+{
+  cached_dir_ensure_loaded (dir);
 
-	  tmp = parent->subdirs;
-	  while (tmp != NULL)
-	    {
-	      CachedDir *subdir = tmp->data;
-
-	      if (!strcmp (subdir->name, dir->name))
-		{
-		  parent->subdirs = g_slist_delete_link (parent->subdirs, tmp);
-		  break;
-		}
-
-	      tmp = tmp->next;
-	    }
-	}
-
-      cached_dir_free (dir);
-    }
-
-  if (parent != NULL)
-    {
-      cached_dir_remove_reference (parent);
-    }
+  return dir->entries;
 }
 
 /*
@@ -724,6 +754,7 @@ entry_directory_new_full (DesktopEntryType  entry_type,
                           const char       *legacy_prefix)
 {
   EntryDirectory *ed;
+  CachedDir      *cd;
   char           *canonical;
 
   menu_verbose ("Loading entry directory \"%s\" (legacy %s)\n",
@@ -738,14 +769,12 @@ entry_directory_new_full (DesktopEntryType  entry_type,
       return NULL;
     }
 
+  cd = cached_dir_load (canonical);
+  g_assert (cd != NULL);
+
   ed = g_new0 (EntryDirectory, 1);
 
-  ed->dir = cached_dir_lookup (canonical);
-  g_assert (ed->dir != NULL);
-
-  cached_dir_add_reference (ed->dir);
-  cached_dir_load_entries_recursive (ed->dir, canonical);
-
+  ed->dir           = cd;
   ed->legacy_prefix = g_strdup (legacy_prefix);
   ed->entry_type    = entry_type;
   ed->is_legacy     = is_legacy != FALSE;
@@ -790,8 +819,6 @@ entry_directory_unref (EntryDirectory *ed)
 
   if (--ed->refcount == 0)
     {
-      cached_dir_remove_reference (ed->dir);
-
       ed->dir        = NULL;
       ed->entry_type = DESKTOP_ENTRY_INVALID;
       ed->is_legacy  = FALSE;
@@ -819,6 +846,42 @@ entry_directory_remove_monitor (EntryDirectory            *ed,
   cached_dir_remove_monitor (ed->dir, ed, callback, user_data);
 }
 
+static DesktopEntry *
+entry_directory_get_desktop (EntryDirectory *ed,
+                             const char     *file_id)
+{
+  DesktopEntry *entry;
+  const char   *unprefixed_id;
+
+  if (ed->entry_type != DESKTOP_ENTRY_DESKTOP)
+    return NULL;
+
+  unprefixed_id = NULL;
+
+  if (ed->is_legacy && ed->legacy_prefix)
+    {
+      if (!g_str_has_prefix (file_id, ed->legacy_prefix))
+        return NULL;
+
+      unprefixed_id = file_id + strlen (ed->legacy_prefix);
+    }
+
+  entry = cached_dir_find_file_id (ed->dir,
+                                   unprefixed_id ? unprefixed_id : file_id,
+                                   ed->is_legacy);
+  if (entry == NULL || desktop_entry_get_type (entry) != DESKTOP_ENTRY_DESKTOP)
+    return NULL;
+
+  if (ed->is_legacy && !desktop_entry_has_categories (entry))
+    {
+      entry = desktop_entry_copy (entry);
+      desktop_entry_add_legacy_category (entry);
+      return entry;
+    }
+
+  return desktop_entry_ref (entry);
+}
+
 static DesktopEntry *
 entry_directory_get_directory (EntryDirectory *ed,
                                const char     *relative_path)
@@ -836,40 +899,30 @@ entry_directory_get_directory (EntryDirectory *ed,
 }
 
 static char *
-get_desktop_file_id_from_path (EntryDirectory   *ed,
-			       DesktopEntryType  entry_type,
-			       const char       *relative_path)
+get_desktop_file_id_from_path (EntryDirectory *ed,
+                               const char     *relative_path)
 {
   char *retval;
-  
-  retval = NULL;
 
-  if (entry_type == DESKTOP_ENTRY_DESKTOP)
+  if (!ed->is_legacy)
     {
-      if (!ed->is_legacy)
-	{
-	  retval = g_strdelimit (g_strdup (relative_path), "/", '-');
-	}
-      else
-	{
-	  char *basename;
-
-	  basename = g_path_get_basename (relative_path);
-
-	  if (ed->legacy_prefix)
-	    {
-	      retval = g_strjoin ("-", ed->legacy_prefix, basename, NULL);
-	      g_free (basename);
-	    }
-	  else
-	    {
-	      retval = basename;
-	    }
-	}
+      retval = g_strdelimit (g_strdup (relative_path), "/", '-');
     }
   else
     {
-      retval = g_strdup (relative_path);
+      char *basename;
+
+      basename = g_path_get_basename (relative_path);
+
+      if (ed->legacy_prefix)
+        {
+          retval = g_strjoin ("-", ed->legacy_prefix, basename, NULL);
+          g_free (basename);
+        }
+      else
+        {
+          retval = basename;
+        }
     }
 
   return retval;
@@ -877,6 +930,7 @@ get_desktop_file_id_from_path (EntryDirectory   *ed,
 
 typedef gboolean (* EntryDirectoryForeachFunc) (EntryDirectory  *ed,
                                                 DesktopEntry    *entry,
+                                                const char      *relative_path,
                                                 const char      *file_id,
                                                 DesktopEntrySet *set,
                                                 gpointer         user_data);
@@ -892,12 +946,9 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
   GSList *tmp;
   int     relative_path_len;
 
-  if (cd->deleted)
-    return TRUE;
-
   relative_path_len = relative_path->len;
 
-  tmp = cd->entries;
+  tmp = cached_dir_get_entries (cd);
   while (tmp != NULL)
     {
       DesktopEntry *entry = tmp->data;
@@ -910,11 +961,11 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
           g_string_append (relative_path,
                            desktop_entry_get_basename (entry));
 
-	  file_id = get_desktop_file_id_from_path (ed,
-						   ed->entry_type,
-						   relative_path->str);
+          file_id = NULL;
+          if (ed->entry_type == DESKTOP_ENTRY_DESKTOP)
+            file_id = get_desktop_file_id_from_path (ed, relative_path->str);
 
-          ret = func (ed, entry, file_id, set, user_data);
+          ret = func (ed, entry, relative_path->str, file_id, set, user_data);
 
           g_free (file_id);
 
@@ -927,13 +978,13 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
       tmp = tmp->next;
     }
 
-  tmp = cd->subdirs;
+  tmp = cached_dir_get_subdirs (cd);
   while (tmp != NULL)
     {
       CachedDir *subdir = tmp->data;
 
       g_string_append (relative_path, subdir->name);
-      g_string_append_c (relative_path, G_DIR_SEPARATOR);
+      g_string_append_c (relative_path, '/');
 
       if (!entry_directory_foreach_recursive (ed,
                                               subdir,
@@ -982,7 +1033,7 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
   if (subdirs)
     *subdirs = NULL;
 
-  tmp = ed->dir->entries;
+  tmp = cached_dir_get_entries (ed->dir);
   while (tmp != NULL)
     {
       DesktopEntry *entry = tmp->data;
@@ -995,9 +1046,7 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
         {
           char *file_id;
 
-          file_id = get_desktop_file_id_from_path (ed,
-						   DESKTOP_ENTRY_DESKTOP,
-						   basename);
+          file_id = get_desktop_file_id_from_path (ed, basename);
 
           desktop_entry_set_add_entry (desktop_entries,
                                        entry,
@@ -1010,8 +1059,8 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
           desktop_entry_get_type (entry) == DESKTOP_ENTRY_DIRECTORY)
         {
           desktop_entry_set_add_entry (directory_entries,
-				       entry,
-				       basename);
+                                       entry,
+                                       g_strdup (basename));
         }
 
       tmp = tmp->next;
@@ -1019,15 +1068,12 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
 
   if (subdirs)
     {
-      tmp = ed->dir->subdirs;
+      tmp = cached_dir_get_subdirs (ed->dir);
       while (tmp != NULL)
         {
           CachedDir *cd = tmp->data;
 
-	  if (!cd->deleted)
-	    {
-	      *subdirs = g_slist_prepend (*subdirs, g_strdup (cd->name));
-	    }
+          *subdirs = g_slist_prepend (*subdirs, g_strdup (cd->name));
 
           tmp = tmp->next;
         }
@@ -1122,6 +1168,25 @@ entry_directory_list_append_list (EntryDirectoryList *list,
   list->dirs = g_list_concat (list->dirs, new_dirs);
 }
 
+DesktopEntry *
+entry_directory_list_get_desktop (EntryDirectoryList *list,
+                                  const char         *file_id)
+{
+  DesktopEntry *retval = NULL;
+  GList        *tmp;
+
+  tmp = list->dirs;
+  while (tmp != NULL)
+    {
+      if ((retval = entry_directory_get_desktop (tmp->data, file_id)) != NULL)
+        break;
+
+      tmp = tmp->next;
+    }
+
+  return retval;
+}
+
 DesktopEntry *
 entry_directory_list_get_directory (EntryDirectoryList *list,
                                     const char         *relative_path)
@@ -1141,74 +1206,13 @@ entry_directory_list_get_directory (EntryDirectoryList *list,
   return retval;
 }
 
-gboolean
-_entry_directory_list_compare (const EntryDirectoryList *a,
-                               const EntryDirectoryList *b)
-{
-  GList *al, *bl;
-
-  if (a == NULL && b == NULL)
-    return TRUE;
-
-  if ((a == NULL || b == NULL))
-    return FALSE;
-
-  if (a->length != b->length)
-    return FALSE;
-
-  al = a->dirs; bl = b->dirs;
-  while (al && bl && al->data == bl->data)
-    {
-      al = al->next;
-      bl = bl->next;
-    }
-
-  return (al == NULL && bl == NULL);
-}
-
-static gboolean
-get_all_func (EntryDirectory   *ed,
-              DesktopEntry     *entry,
-              const char       *file_id,
-              DesktopEntrySet  *set,
-              gpointer          user_data)
-{
-  if (ed->is_legacy && !desktop_entry_has_categories (entry))
-    {
-      entry = desktop_entry_copy (entry);
-      desktop_entry_add_legacy_category (entry);
-    }
-  else
-    {
-      entry = desktop_entry_ref (entry);
-    }
-
-  desktop_entry_set_add_entry (set, entry, file_id);
-  desktop_entry_unref (entry);
-
-  return TRUE;
-}
-
-static DesktopEntrySet    *entry_directory_last_set = NULL;
-static EntryDirectoryList *entry_directory_last_list = NULL;
-
-void
-_entry_directory_list_empty_desktop_cache (void)
-{
-  if (entry_directory_last_set != NULL)
-    desktop_entry_set_unref (entry_directory_last_set);
-  entry_directory_last_set = NULL;
-
-  if (entry_directory_last_list != NULL)
-    entry_directory_list_unref (entry_directory_last_list);
-  entry_directory_last_list = NULL;
-}
-
-DesktopEntrySet *
-_entry_directory_list_get_all_desktops (EntryDirectoryList *list)
+static void
+entry_directory_list_add (EntryDirectoryList        *list,
+                          EntryDirectoryForeachFunc  func,
+                          DesktopEntrySet           *set,
+                          gpointer                   user_data)
 {
   GList *tmp;
-  DesktopEntrySet *set;
 
   /* The only tricky thing here is that desktop files later
    * in the search list with the same relative path
@@ -1223,35 +1227,150 @@ _entry_directory_list_get_all_desktops (EntryDirectoryList *list)
    * entry)
    */
 
-  /* This method is -extremely- slow, so we have a simple
-     one-entry cache here */
-  if (_entry_directory_list_compare (list, entry_directory_last_list))
-    {
-      menu_verbose (" Hit desktop list (%p) cache\n", list);
-      return desktop_entry_set_ref (entry_directory_last_set);
-    }
-
-  if (entry_directory_last_set != NULL)
-    desktop_entry_set_unref (entry_directory_last_set);
-  if (entry_directory_last_list != NULL)
-    entry_directory_list_unref (entry_directory_last_list);
-
-  set = desktop_entry_set_new ();
-  menu_verbose (" Storing all of list %p in set %p\n",
-                list, set);
-
   tmp = g_list_last (list->dirs);
   while (tmp != NULL)
     {
-      entry_directory_foreach (tmp->data, get_all_func, set, NULL);
+      entry_directory_foreach (tmp->data, func, set, user_data);
 
       tmp = tmp->prev;
     }
+}
 
-  entry_directory_last_list = entry_directory_list_ref (list);
-  entry_directory_last_set = desktop_entry_set_ref (set);
+static gboolean
+get_all_func (EntryDirectory   *ed,
+              DesktopEntry     *entry,
+              const char       *relative_path,
+              const char       *file_id,
+              DesktopEntrySet  *set,
+              gpointer          user_data)
+{
+  if (ed->is_legacy && !desktop_entry_has_categories (entry))
+    {
+      entry = desktop_entry_copy (entry);
+      desktop_entry_add_legacy_category (entry);
+    }
+  else
+    {
+      entry = desktop_entry_ref (entry);
+    }
 
-  return set;
+  desktop_entry_set_add_entry (set,
+                               entry,
+                               file_id ? file_id : relative_path);
+  desktop_entry_unref (entry);
+
+  return TRUE;
+}
+
+void
+entry_directory_list_get_all_desktops (EntryDirectoryList *list,
+                                       DesktopEntrySet    *set)
+{
+  menu_verbose (" Storing all of list %p in set %p\n",
+                list, set);
+
+  entry_directory_list_add (list, get_all_func, set, NULL);
+}
+
+static gboolean
+get_by_category_func (EntryDirectory  *ed,
+                      DesktopEntry    *entry,
+                      const char      *relative_path,
+                      const char      *file_id,
+                      DesktopEntrySet *set,
+                      gpointer         user_data)
+{
+  const char *category = user_data;
+
+  if (ed->is_legacy && !desktop_entry_has_categories (entry))
+    {
+      if (strcmp (category, "Legacy") != 0)
+        return TRUE;
+
+      entry = desktop_entry_copy (entry);
+      desktop_entry_add_legacy_category (entry);
+    }
+  else
+    {
+      if (!desktop_entry_has_category (entry, category))
+        return TRUE;
+
+      entry = desktop_entry_ref (entry);
+    }
+
+  desktop_entry_set_add_entry (set,
+                               entry,
+                               file_id ? file_id : relative_path);
+  desktop_entry_unref (entry);
+
+  return TRUE;
+}
+
+void
+entry_directory_list_get_by_category (EntryDirectoryList *list,
+                                      const char         *category,
+                                      DesktopEntrySet    *set)
+{
+  menu_verbose (" Storing list %p category %s in set %p\n",
+                list, category, set);
+
+  entry_directory_list_add (list,
+                            get_by_category_func,
+                            set,
+                            (char *) category);
+}
+
+static gboolean
+get_inverse_func (EntryDirectory  *ed,
+                  DesktopEntry    *entry,
+                  const char      *relative_path,
+                  const char      *file_id,
+                  DesktopEntrySet *set,
+                  gpointer         user_data)
+{
+  DesktopEntrySet *inverse = user_data;
+
+  /* if not in the original set, add to inverse set */
+
+  if (desktop_entry_set_lookup (set, relative_path) != NULL)
+    return TRUE;
+
+  if (ed->is_legacy && !desktop_entry_has_categories (entry))
+    {
+      entry = desktop_entry_copy (entry);
+      desktop_entry_add_legacy_category (entry);
+    }
+  else
+    {
+      entry = desktop_entry_ref (entry);
+    }
+
+  desktop_entry_set_add_entry (inverse,
+                               entry,
+                               file_id ? file_id : relative_path);
+  desktop_entry_unref (entry);
+
+  return TRUE;
+}
+
+void
+entry_directory_list_invert_set (EntryDirectoryList *list,
+                                 DesktopEntrySet    *set)
+{
+  DesktopEntrySet *inverse;
+
+  menu_verbose (" Inverting set %p relative to list %p\n",
+                set, list);
+
+  inverse = desktop_entry_set_new ();
+  entry_directory_list_add (list,
+                            get_inverse_func,
+                            set,
+                            inverse);
+
+  desktop_entry_set_swap_contents (set, inverse);
+
+  desktop_entry_set_unref (inverse);
 }
 
 void
